/**
  Physics simulator demo program.
  
  Dr. Orion Sky Lawlor, olawlor@acm.org, 2011-01-20 (Public Domain)
*/
#include "physics/world.h"
#include "ogl/glsl.h"
#include "soil/SOIL.h"

/* LAME! just include library bodies here, for easy linking */
#include "physics/world.cpp" 
#include "physics/config.cpp"
#include "ogl/glsl.cpp"
#include "soil/SOIL.c"
#include "soil/stb_image_aug.c"


/* One simulated particle */
class particle {
public:
	vec4 color; // RGBA
	vec3 pos, vel; // in world coordinates-- meters
	float age; // time in seconds since created
	bool flip; // true if our texcoords should be flipped backwards
	
	void simulate(double dt) {
	/* Newtonian motion */
		vec3 acc(0,0,-9.8); 
		pos+=dt*vel;
		vel+=dt*acc;
		age+=dt;
		if (age>2.5) color-=vec4(dt); // darken before vanishing
		
	/* Boundary conditions */
		if (pos.z<0.0) { /* hit the ground */
			vel.z=fabs(vel.z)-0.2;
			
			vec3 F=-0.1*vel;
			vec3 A=F/1.0;
			vel+=dt*A;
			
			pos.z=0.0;
		}	
		
		float r=6.0;
		vec3 center=vec3(4.0,12.0,0);
		if (length(pos-center)<r) { /* hit big sphere */
			vel=0.7*reflect(vel,normalize(pos-center));
			pos=normalize(pos-center)*r+center;
		}
	}
};

/* Random float between 0 and 1 */
float randf(void) {
	float f=rand();
	return f/RAND_MAX;
}

/* Random vector where x y and z vary between 0 and 1 */
vec3 randVec3(void) {
	return vec3(randf(),randf(),randf());
}

/* Clean up shader (back to ordinary OpenGL) */ 
void endPointShader(physics::library &lib) {
	if (!lib.hasShaders) return;
	glUseProgramObjectARB(0);
}

/* A particle system */
class particlesystem : public physics::object {
	std::vector<particle> p;
	float max_age; // seconds to keep particles around
	
	particle make_particle(float max_age) {
		particle np;
		np.color=randVec3(); np.color.w=0.5; /* alpha */
		np.pos=randVec3()*0.3+vec3(0.0,0.0,1.4); /* push up */
		np.vel=vec3(2.7,6.8,5.0); // (randVec3()-vec3(0.5));
		np.flip=(rand()%2)==1; // flip half of 'em backwards
		np.age=randf()*max_age;
		return np;
	}
	
public:
	particlesystem(void) 
		:physics::object(0.01)  /* <- our timestep, in seconds */
	{
		max_age=3.5;
		for (int i=0;i<10000;i++) {
			p.push_back(make_particle(max_age));
		}
	}
	
	void simulate(physics::library &lib) {
		float slowmo=0.7;
		for (unsigned int i=0;i<p.size();i++) {
			p[i].simulate(timestep.dt*slowmo);
			if (p[i].age>max_age) { // replace old particle with a fresh one
				p[i]=make_particle(0.0);
			}
		}
	}
	
	void draw(physics::library &lib) {
		/* Draw a teapot at the emission point */
		glColor3f(1.0,0.3,0.1);
		glPushMatrix();
		glTranslatef(-0.5,-1.7,1.0);
		glRotatef(70,0.0,0.0,1.0); // orient spout along stream
		glRotatef(90,1.0,0.0,0.0); // GLUT teapot has Y==up (infidels!)
		glutWireTeapot(1.3);
		glPopMatrix();
		glFlush();
		
		/* Load up fire texture with SOIL */
		static GLuint texture=SOIL_load_OGL_texture(
			"fire.png",0,0,SOIL_FLAG_MIPMAPS
		);
		glBindTexture(GL_TEXTURE_2D,texture);
		glEnable(GL_TEXTURE_2D);

		vec3 corners[4], texcoords[4], flipcoords[4];
		float size=0.05; // half the size of our little billboard
		vec3 screenx=camera_orient.x*size;  // billboards MUST face the camera!
		vec3 screeny=camera_orient.y*size;
		corners[0]=-1.0*screenx+1.0*screeny; corners[1]=screenx+1.0*screeny;
		corners[3]=-1.0*screenx-1.0*screeny; corners[2]=screenx-1.0*screeny;
		texcoords[0]=vec3(0,0,0); texcoords[1]=vec3(1,0,0); 
		texcoords[3]=vec3(0,1,0); texcoords[2]=vec3(1,1,0); 
		flipcoords[0]=vec3(1,0,0); flipcoords[1]=vec3(0,0,0); 
		flipcoords[3]=vec3(1,1,0); flipcoords[2]=vec3(0,1,0); 
		
		
		glMatrixMode(GL_TEXTURE); // slowly rotate the fire texture
		glLoadIdentity();
		glTranslatef(0.5,0.5,0.0);
		glRotatef(lib.time*100.0f,0.0f,0.0f,1.0f);
		glTranslatef(-0.5,-0.5,0.0);
		
	/* FIRE needs additive-only alpha blending-- this can be done in any order! */
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA,GL_ONE); // _MINUS_SRC_ALPHA);
		glDepthMask(GL_FALSE); // they're transparent, so no depth writes
		
		glBegin(GL_QUADS);
		for (unsigned int i=0;i<p.size();i++) {
			glColor4fv(p[i].color);
			for (int corner=0;corner<4;corner++) {
				if (p[i].flip)
					glTexCoord2fv(flipcoords[corner]);
				else
					glTexCoord2fv(texcoords[corner]);
				glVertex3fv(p[i].pos+corners[corner]);
			}
		}
		glEnd();
		glDepthMask(GL_TRUE); // reset OpenGL state
		glDisable(GL_TEXTURE_2D);
	}
};


/* Called to create a new simulation */
void physics_setup(physics::library &lib) {
	lib.world->add(new particlesystem());
	lib.background[0]=lib.background[1]=lib.background[2]=0.3; /* dark grey */
	camera=vec3(1.5,-3,1.7);
}

/* Called every frame to draw everything */
void physics_draw_frame(physics::library &lib) {
	/* Point and line smoothing needs normal alpha blending */
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	
	/* No depth while drawing grid */
	glDisable(GL_DEPTH_TEST);
	glColor3f(0.0,0.0,0.0);
	glEnable(GL_LINE_SMOOTH);
	glLineWidth(0.2);
	if (camera.z>=0.0) { /* no grid if you're underneath ground */
		glBegin(GL_LINES);
		for (int grid=-100;grid<=+100;grid++) {
			glVertex3i(-100,grid,0); glVertex3i(+100,grid,0); 
			glVertex3i(grid,-100,0); glVertex3i(grid,+100,0); 
		}
		glEnd();
	}
	
	/* Do depth while drawing objects */
	glEnable(GL_DEPTH_TEST);
	lib.world->draw(lib);
}


